
var _ = require('underscore'),
    $ = require('ore').query,
    html = require('ore/html'),
    webgl = require('./gl'),
    shaders = require('./shaders'),
    glMatrix = require('./glMatrix');

var vec3 = glMatrix.vec3;
var mat4 = glMatrix.mat4;

exports.Canvas3D = html.canvas('.Canvas3D', [],
{
    fps: 60,
    
    construct: function() {
        this.models = [];
        this.lights = [];
        
        _.each(this.scene, _.bind(function(object) {
            if (object instanceof Model) {
                this.models[this.models.length] = object;
            } else if (object instanceof Light) {
                this.lights[this.lights.length] = object;
            } else if (object instanceof Camera) {
                this.camera = object;
            }
        }, this));
        
        if (!this.camera) {
            this.camera = camera();
        }

        this.interval = setTimeout(_.bind(this.draw, this), 1000 / this.fps);
        // this.interval = setInterval(_.bind(this.draw, this), 1000 / this.fps);
    },

    draw: function() {
        this._syncSize();
        
        var gl = this.gl;
        gl.viewport(0, 0, this.viewportWidth, this.viewportHeight);
        gl.clearColor(0.0, 0.0, 0.0, 1.0);
        gl.clearDepth(1.0);
        gl.enable(gl.DEPTH_TEST);
        gl.depthFunc(gl.LEQUAL);
        gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
        
        var camera = this.camera;
        var lights = this.lights;
        
        var perspective = camera.perspective(this);
        var p = new Float32Array(perspective);

        _.each(this.models, function(model) {
            model.material.createProgram(gl);
            var program = model.material.program;
            if (program) {
                gl.useProgram(program);

                var mv = model.transform();
                var n = mat4.transpose(mat4.inverse(model.transform(), mat4.create()));
                
                var uniforms = program.uniforms;
                gl.uniformMatrix4fv(uniforms.uPMatrix.uniform, false, p);
                gl.uniformMatrix4fv(uniforms.uMVMatrix.uniform, false, mv);
                gl.uniformMatrix4fv(uniforms.uNMatrix.uniform, false, n);
                gl.uniform1i(uniforms.uPerFragmentLighting.uniform, 1);

                _.each(lights, function(light) {
                    if (light.isAmbient) {
                        gl.uniform3f(uniforms.uAmbientColor.uniform, light.r, light.g, light.b);
                    } else {
                        gl.uniform3f(uniforms.uLightColor.uniform, light.r, light.g, light.b);
                        gl.uniform3fv(uniforms.uLightPosition.uniform, [light.x, light.y, light.z]);
                        gl.uniform3fv(uniforms.uLightDirection.uniform, light.direction());
                    }
                });

                if (uniforms.uImage) {
                    gl.activeTexture(gl.TEXTURE0);
                    gl.bindTexture(gl.TEXTURE_2D, model.material.texture);
                    gl.uniform1i(uniforms.uImage.uniform, 0);
                }

                model.createBuffers(gl);

                gl.bindBuffer(gl.ARRAY_BUFFER, model.vertexBuffer);
                gl.vertexAttribPointer(program.aVertexPosition, 3, gl.FLOAT, false, 0, 0);

                gl.bindBuffer(gl.ARRAY_BUFFER, model.normalsBuffer);
                gl.vertexAttribPointer(program.aVertexNormal, 3, gl.FLOAT, false, 0, 0);

                gl.bindBuffer(gl.ARRAY_BUFFER, model.texcoordBuffer);
                gl.vertexAttribPointer(program.aTextureCoord, 2, gl.FLOAT, false, 0, 0);

                gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, model.indexBuffer);
                gl.drawElements(gl.TRIANGLES, model.indexBuffer.length, gl.UNSIGNED_SHORT, 0);
            }
        });
    },
    
    _getFramebuffer: function(name, width, height) {
        var key = name + ':' + width + ':' + height;
        if (!(key in this.framebuffers)) {
            this.framebuffers[key] = this._createFramebuffer(width, height);
        }
        return this.framebuffers[key];
    },
    
    _createFramebuffer: function(width, height) {
        var gl = this.gl;
        var framebuffer = gl.createFramebuffer();
        gl.bindFramebuffer(gl.FRAMEBUFFER, framebuffer);
        framebuffer.width = width;
        framebuffer.height = height;
    
        var texture = gl.createTexture();
        texture.width = width;
        texture.height = height;
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR_MIPMAP_NEAREST);
        gl.generateMipmap(gl.TEXTURE_2D);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, width, height, 0, gl.RGBA, gl.UNSIGNED_BYTE, null);
    
        var renderbuffer = gl.createRenderbuffer();
        gl.bindRenderbuffer(gl.RENDERBUFFER, renderbuffer);
        gl.renderbufferStorage(gl.RENDERBUFFER, gl.DEPTH_COMPONENT16, width, height);
    
        gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, texture, 0);
        gl.framebufferRenderbuffer(gl.FRAMEBUFFER, gl.DEPTH_ATTACHMENT, gl.RENDERBUFFER,
                                   renderbuffer);
    
        gl.bindTexture(gl.TEXTURE_2D, null);
        gl.bindRenderbuffer(gl.RENDERBUFFER, null);
        gl.bindFramebuffer(gl.FRAMEBUFFER, null);

        return {
            framebuffer: framebuffer,
            texture: texture,
        };
    },
    
    _syncSize: function() {
        var canvas = this.val();
        if (!this.gl
            || canvas.width != canvas.offsetWidth || canvas.height != canvas.offsetHeight) {
            canvas.width = canvas.offsetWidth;
            canvas.height = canvas.offsetHeight;
            this.viewportWidth = canvas.width;
            this.viewportHeight = canvas.height;
            
            this.gl = canvas.getContext('experimental-webgl');
        }
    },
    
    _drawShaders: function(width, height, textures, programs) {
        // Get framebuffers if we need them for offscreen rendering
        var buffer1 = this.programs.length > 1 ? this._getFramebuffer('a', width, height) : null;
        var buffer2 = this.programs.length > 2 ? this._getFramebuffer('b', width, height) : null;

        // Draw each of the intermediate shaders to offscreen textures
        var buffer = buffer1;
        for (var i = 0; i < programs.length-1; ++i) {
            this._drawImage(texture, programs[i], buffer.framebuffer);
            texture = buffer.texture;
            buffer = buffer == buffer1 ? buffer2 : this.buffer1;
        }
        return texture;
    },

    _drawImage: function(texture, program, framebuffer) {
        var gl = this.gl;

        if (framebuffer) {
            gl.bindFramebuffer(gl.FRAMEBUFFER, framebuffer);
        }
        
        gl.viewport(0, 0, texture.width, texture.height);
        gl.clear(gl.COLOR_BUFFER_BIT);

        gl.useProgram(program);

        _.each(program.uniforms, function(value, name) {
            value.setter(gl, value.uniform, value.map[name]);
        });

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.uniform1i(program.uImage, 0);

        // XXXjoe Create the plane*Buffers at some point
        
        gl.bindBuffer(gl.ARRAY_BUFFER, this.planeVertexBuffer);
        gl.vertexAttribPointer(program.aVertexPosition, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.planeTexcoordBuffer);
        gl.vertexAttribPointer(program.aTextureCoord, 2, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.planeIndexBuffer);
        gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0);

        if (framebuffer) {
            gl.bindFramebuffer(gl.FRAMEBUFFER, null);
        }
    },
});

// *************************************************************************************************

function Model(material, vertices, normals, texcoords, indices, matrix) {
    this.setMaterial(material);
    this.vertices = vertices;
    this.normals = normals;
    this.texcoords = texcoords;
    this.indices = indices;
    this.matrix = matrix;
}
Canvas3D.Model = Model;

Model.prototype = {
    clone: function() {
        return new Model(this.material, this.getVertices(), this.getNormals(),
                         this.getTexcoords(), this.getIndices(), this.transform());
    },
    
    setMaterial: function(material) {
        if (material instanceof Material) {
            this.material = material;
        } else {
            this.material = new Material(material);
        }
    },
    
    getVertices: function() {
        return this.vertices || [];
    },
    
    getNormals: function() {
        return this.normals || [];
    },
    
    getTexcoords: function() {
        return this.texcoords || [];
    },
    
    getIndices: function() {
        return this.indices || [];
    },
    
    createBuffers: function(gl) {
        if (!this.vertexBuffer) {
            var vertices = this.getVertices();
            this.vertexBuffer = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, this.vertexBuffer);
            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertices), gl.STATIC_DRAW);
        }
        
        if (!this.normalsBuffer) {
            var normals = this.getNormals();
            this.normalsBuffer = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, this.normalsBuffer);
            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(normals), gl.STATIC_DRAW);
        }
        
        if (!this.texcoordBuffer) {
            var texcoords = this.getTexcoords();
            this.texcoordBuffer = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, this.texcoordBuffer);
            gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(texcoords), gl.STATIC_DRAW);
        }
    
        if (!this.indexBuffer) {                
            var indices = this.getIndices();
            this.indexBuffer = gl.createBuffer();
            gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indexBuffer);
            gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(indices), gl.STATIC_DRAW);
            this.indexBuffer.length = indices.length;
        }
    },
        
    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
    
    transform: function() {
        if (!this.matrix) {
            this.matrix = mat4.identity(mat4.create());
        }
        return this.matrix;
    },

    translate: function(x, y, z) {
        mat4.translate(this.transform(), [x, y, z]);
        return this;
    },

    rotate: function(angle, x, y, z) {
        var radians = angle * (Math.PI / 180.0);
        mat4.rotate(this.transform(), radians, [x, y, z]);
        return this;
    },

    scale: function(x, y, z) {
        mat4.scale(this.transform(), [x, y, z]);
        return this;
    },
};

// *************************************************************************************************

function Material(fragments, vertex, src) {
    this.fragmentShaders = (fragments && !$.isArray(fragments) ? [fragments] : fragments)
                           || [shaders.normal];
    this.vertexShader = vertex || shaders.perspectiveVertexShader;
    this.src = src;
}
Canvas3D.Material = Material;

Material.prototype = {
    createProgram: function(gl) {
        if (!this.loading && !this.texture && this.src) {
            this._loadTexture(this.src, gl);
        } else if (!this.program) {
            this.program = initProgram(gl, this.vertexShader, this.fragmentShaders[0]);
        }
    },
    
    _loadTexture: function(url, gl) {
        this.loading = true;

        var image = new Image();
        image.onload = _.bind(function() {
            var texture = gl.createTexture();
            texture.width = image.width;
            texture.height = image.height;
            gl.bindTexture(gl.TEXTURE_2D, texture);
            gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, true);
            gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
            gl.bindTexture(gl.TEXTURE_2D, null);
            
            this.texture = texture;
            this.loading = false;
        }, this);
        image.onerror = _.bind(function() {
            this.loading = false;
        }, this);
        image.src = url;
    }
};

// *************************************************************************************************

function camera(fov, zMin, zMax) {
    return new Camera(fov, zMin, zMax);
}
Canvas3D.camera = camera;

function Camera(fov, zMin, zMax) {    
    this.fov = fov || 45;
    this.zMin = zMin || 0.1;
    this.zMax = zMax || 100;
}

Camera.prototype = {
    fov: 45,
    zMin: 0.1,
    zMax: 100,
    aspect: 0,
    yaw: 0,
    pitch: 0,
    x: 0,
    y: 0,
    z: 5,
    
    translate: function(x, y, z) {
        mat4.translate(this.view(), [x, y, z]);
        return this;
    },

    rotate: function(angle, x, y, z) {
        var radians = angle * (Math.PI / 180.0);
        mat4.rotate(this.view(), radians, [x, y, z]);
        return this;
    },

    view: function() {
        if (!this.matrix) {
            this.matrix = mat4.identity(mat4.create());
        }
        return this.matrix;
    },

    perspective: function(canvas) {
        var aspect = this.aspect || (canvas.viewportWidth/canvas.viewportHeight);
        var p = mat4.perspective(this.fov, aspect, this.zMin, this.zMax);
        return mat4.multiply(p, this.view());
    },
};

// *************************************************************************************************

function light(r, g, b, isAmbient) {
    return new Light(r, g, b, isAmbient);
}
Canvas3D.light = light;

function Light(r, g, b, isAmbient) {
    this.r = r;
    this.g = g;
    this.b = b;
    this.isAmbient = isAmbient;
}

Light.prototype = {
    x: 0,
    y: 0,
    z: 0,
    isAmbient: false,
    
    translate: function(x, y, z) {
        this.x += x;
        this.y += y;
        this.z += z;
        return this;
    },

    direction: function() {
        var direction = vec3.create([this.x, this.y, this.z]);
        return vec3.negate(vec3.normalize(direction));
    }
};

// *************************************************************************************************

function initProgram(gl, vertexSource, fragmentSource) {
    var program = gl.createProgram();
    
    var vertex = compileShader(gl, vertexSource, gl.VERTEX_SHADER);
    if (vertex) {
        gl.attachShader(program, vertex);
    }

    var fragment = compileShader(gl, fragmentSource, gl.FRAGMENT_SHADER);
    if (fragment) {
        gl.attachShader(program, fragment);
    }

    gl.linkProgram(program);
    if (gl.getProgramParameter(program, gl.LINK_STATUS)) {
        program.aVertexPosition = gl.getAttribLocation(program, "aVertexPosition");
        gl.enableVertexAttribArray(program.aVertexPosition);

        program.aVertexNormal = gl.getAttribLocation(program, "aVertexNormal");
        gl.enableVertexAttribArray(program.aVertexNormal);
        
        program.aTextureCoord = gl.getAttribLocation(program, "aTextureCoord");
        gl.enableVertexAttribArray(program.aTextureCoord);

        program.uniforms = getUniformMap(gl, program);
        return program;
    } else {
        console.error(gl.getShaderInfoLog(shader));
    }
}

function compileShader(gl, source, type) {
    if (typeof(source) == 'function') {
        source = source();
    }
    if (typeof(source) == 'object') {
        source = source.fn+'';
        source = source.replace(/^function \(\) \{\s*\/\*/, '').replace(/\*\/\s*;?\s*}/, '');
    }
    
    if (type == gl.FRAGMENT_SHADER) {
        source = webgl.defaultShaderHeader + source;
    }
    
    var shader = gl.createShader(type);
    gl.shaderSource(shader, source);
    gl.compileShader(shader);
    if (gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
        return shader;
    } else {
        console.error(gl.getShaderInfoLog(shader));
    }
}

function getUniformMap(gl, program) {
    var uniforms = {};
    var count = gl.getProgramParameter(program, gl.ACTIVE_UNIFORMS);
    for (var i = 0; i < count; ++i) {
        var info = gl.getActiveUniform(program, i);
        var uniform = gl.getUniformLocation(program, info.name);
        var setter = webgl.uniformMap[info.type];
        uniforms[info.name] = {uniform:uniform, name: info.name, setter:setter};
        if (!setter) {
            console.warn('Type', info.type, 'unknown for', info.name);
        }
    }
    return uniforms;
}
