
var _ = require('underscore'),
    $ = require('ore').query,
    html = require('ore/html'),
    webgl = require('./gl'),
    shaders = require('./shaders');

exports.Picture = html.canvas('.Picture', [],
{
    construct: function() {
        setTimeout(_.bind(function() {
            this.loadImage();
        }, this));
    },
    
    shader: function(id) {
        return $.detect(this.shaders, function(shader) { return shader.id == id; });
    },
    
    draw: function() {
        if (this.programs.length > 1 && !this.buffer1) {
            this.buffer1 = this.createFramebuffer();
        }
        if (this.programs.length > 2 && !this.buffer2) {
            this.buffer2 = this.createFramebuffer();
        }
        
        // Draw each of the intermediate shaders to offscreen textures
        var texture = this.texture;
        var buffer = this.buffer1;
        for (var i = 0; i < this.programs.length-1; ++i) {
            this.drawImage(texture, this.programs[i], buffer);
            texture = buffer.texture;
            buffer = buffer == this.buffer1 ? this.buffer2 : this.buffer1;
        }

        // Draw the final texture to the screen
        this.drawImage(texture, this.programs[this.programs.length-1], null);
    },
    
    drawImage: function(texture, program, buffer) {
        var gl = this.gl;

        if (buffer) {
            gl.bindFramebuffer(gl.FRAMEBUFFER, buffer.framebuffer);
        }
        
        gl.viewport(0, 0, this.width, this.height);
        gl.clear(gl.COLOR_BUFFER_BIT);

        gl.useProgram(program);

        _/each(program.uniforms, function(value, name) {
            value.setter(gl, value.uniform, value.map[name]);
        });

        gl.activeTexture(gl.TEXTURE0);
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.uniform1i(program.uImage, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.vertexBuffer);
        gl.vertexAttribPointer(program.aVertexPosition, 3, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ARRAY_BUFFER, this.textureCoordBuffer);
        gl.vertexAttribPointer(program.aTextureCoord, 2, gl.FLOAT, false, 0, 0);

        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indexBuffer);
        gl.drawElements(gl.TRIANGLES, 6, gl.UNSIGNED_SHORT, 0);

        if (buffer) {
            gl.bindFramebuffer(gl.FRAMEBUFFER, null);
        }
    },
        
    initCanvas: function() {
        var canvas = this.val();
        canvas.width = canvas.offsetWidth;
        canvas.height = canvas.offsetHeight;
        this.gl = canvas.getContext('experimental-webgl');
    },

    initPrograms: function() {
        this.programs = [];
        if (!this.shaders || !this.shaders.length) {
            this.shaders = [shaders.normal];
        }
        for (var i = 0; i < this.shaders.length; ++i) {
            var program = this.createProgram(this.shaders[i], i);
            if (program) {
                this.programs.push(program);
            }
        }
    },

    initBuffers: function() {
        var gl = this.gl;
        
        var vertices = [
          -1.0, -1.0,  1.0,
           1.0, -1.0,  1.0,
           1.0,  1.0,  1.0,
          -1.0,  1.0,  1.0,
        ];
        this.vertexBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.vertexBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(vertices), gl.STATIC_DRAW);

        var textureCoords = [
          0.0, 0.0,
          1.0, 0.0,
          1.0, 1.0,
          0.0, 1.0,
        ];
        this.textureCoordBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ARRAY_BUFFER, this.textureCoordBuffer);
        gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(textureCoords), gl.STATIC_DRAW);

        var vertexIndices = [
          0, 1, 2, 0, 2, 3,
        ];
        this.indexBuffer = gl.createBuffer();
        gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.indexBuffer);
        gl.bufferData(gl.ELEMENT_ARRAY_BUFFER, new Uint16Array(vertexIndices), gl.STATIC_DRAW);
    },

    initTexture: function(image) {
        image = webgl.createPowerOfTwoImage(image);

        var gl = this.gl;
        var texture = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.pixelStorei(gl.UNPACK_FLIP_Y_WEBGL, true);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.MIRRORED_REPEAT);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.MIRRORED_REPEAT);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image);
        gl.bindTexture(gl.TEXTURE_2D, null);
        this.texture = texture;
    },

    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
        
    loadImage: function() {
        this.loadTexture(this.src, _.bind(function(image) {
            var canvas = this.val();
            
            var s = getComputedStyle(canvas);
            if (s.width == '0px') {
                this.css('width', image.width);
            }
            if (s.height == '0px') {
                this.css('height', image.height);
            }

            this.width = webgl.nextHighestPowerOfTwo(image.width);
            this.height = webgl.nextHighestPowerOfTwo(image.height);

            this.initCanvas();
            this.initPrograms();
            this.initBuffers();
            this.initTexture(image);
            this.draw();
        }, this));
    },
    
    loadTexture: function(url, cb) {
        var image = new Image();
        image.onload = _.bind(function() {
            cb(image);
        }, this);
        image.onerror = _.bind(function() {
            cb(null);
        }, this);
        image.src = url;
    },
    
    createProgram: function(source, i) {
        var gl = this.gl;
        var fragmentShader = this.compileShader(source, gl.FRAGMENT_SHADER);
        if (fragmentShader) {
            var program = this.program = gl.createProgram();
            var vertexShader = this.compileShader(shaders.flatVertexShader, gl.VERTEX_SHADER);
            gl.attachShader(program, vertexShader);
            gl.attachShader(program, fragmentShader);
        }

        gl.linkProgram(program);
        if (gl.getProgramParameter(program, gl.LINK_STATUS)) {
            program.aVertexPosition = gl.getAttribLocation(program, "aVertexPosition");
            gl.enableVertexAttribArray(program.aVertexPosition);
            
            program.aTextureCoord = gl.getAttribLocation(program, "aTextureCoord");
            gl.enableVertexAttribArray(program.aTextureCoord);
            
            program.uImage = gl.getUniformLocation(program, "uImage");

            if (typeof(source) == 'object') {
                var uniforms = {};
                var count = gl.getProgramParameter(program, gl.ACTIVE_UNIFORMS);
                for (var i = 0; i < count; ++i) {
                    var info = gl.getActiveUniform(program, i);
                    if (info.name in source.uniforms) {
                        var uniform = gl.getUniformLocation(program, info.name);
                        var setter = webgl.uniformMap[info.type];
                        if (setter) {
                            uniforms[info.name] = {uniform:uniform, name: info.name,
                                                   map: source.uniforms, setter:setter};
                        } else {
                            console.warn('Type', info.type, 'unknown for', info.name);
                        }
                    }
                }
                program.uniforms = uniforms;
            }
                        
            return program;
        }
    },

    createFramebuffer: function() {
        var gl = this.gl;
        var framebuffer = gl.createFramebuffer();
        gl.bindFramebuffer(gl.FRAMEBUFFER, framebuffer);
        framebuffer.width = this.width;
        framebuffer.height = this.height;

        var texture = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.MIRRORED_REPEAT);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.MIRRORED_REPEAT);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, this.width, this.height, 0, gl.RGBA,
                      gl.UNSIGNED_BYTE, null);
        gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0, gl.TEXTURE_2D, texture, 0);
        gl.bindTexture(gl.TEXTURE_2D, null);

        gl.bindFramebuffer(gl.FRAMEBUFFER, null);
        
        return {
            framebuffer: framebuffer,
            texture: texture,
        };
    },
    
    compileShader: function(source, type) {
        if (typeof(source) == 'function') {
            source = source();
        }
        if (typeof(source) == 'object') {
            source = source.fn+'';
            source = source.replace(/^function \(\) \{\s*\/\*/, '').replace(/\*\/\s*;?\s*}/, '');
        }
        
        if (type == this.gl.FRAGMENT_SHADER) {
            source = webgl.defaultShaderHeader + source;
        }
        
        var shader = this.gl.createShader(type);
        this.gl.shaderSource(shader, source);
        this.gl.compileShader(shader);
        if (this.gl.getShaderParameter(shader, this.gl.COMPILE_STATUS)) {
            return shader;
        } else {
            console.error(this.gl.getShaderInfoLog(shader));
        }
    },

});
