
var webgl = require('./gl');

var shader = webgl.shader;

// *************************************************************************************************
// Vertex shaders

exports.flatVertexShader = shader(function() {/*

attribute vec3 aVertexPosition;
attribute vec2 aTextureCoord;
varying vec2 vTextureCoord;

void main() {
    gl_Position = vec4(aVertexPosition, 1);
    vTextureCoord = aTextureCoord;
}

*/});

exports.perspectiveVertexShader = shader(function() {/*

attribute vec3 aVertexPosition;
attribute vec3 aVertexNormal;
attribute vec2 aTextureCoord;

uniform mat4 uMVMatrix;
uniform mat4 uPMatrix;
uniform mat4 uNMatrix;

uniform vec3 uAmbientColor;
uniform vec3 uLightDirection;
uniform vec3 uLightPosition;
uniform vec3 uLightColor;
uniform bool uPerFragmentLighting;

varying vec2 vTextureCoord;

varying vec3 vLightWeighting;
varying vec4 vTransformedNormal;
varying vec4 vPosition;

void main() {
    vPosition = uMVMatrix * vec4(aVertexPosition, 1);
    gl_Position = uPMatrix * vPosition;

    vTextureCoord = aTextureCoord;

    vTransformedNormal = uNMatrix * vec4(aVertexNormal, 1.0);

    if (uPerFragmentLighting) {
        vec4 transformedNormal = uNMatrix * vec4(aVertexNormal, 1.0);
        float directionalLightWeighting = max(dot(transformedNormal.xyz, uLightDirection), 0.0);
        vLightWeighting = uAmbientColor + uLightColor * directionalLightWeighting;
    } else {
        vec3 lightDirection = normalize(uLightPosition - vPosition.xyz);
        float directionalLightWeighting =
            max(dot(normalize(vTransformedNormal.xyz), lightDirection), 0.0);
        vLightWeighting = uAmbientColor + uLightColor * directionalLightWeighting;
        // vLightWeighting = vec3(1.0, 1.0, 1.0);
    }

}

*/});

// *************************************************************************************************
// Fragment shaders

exports.normal = shader(function() {/*

varying vec2 vTextureCoord;
uniform sampler2D uImage;

void main(void) {
    gl_FragColor = texture2D(uImage, vTextureCoord);
}

*/});

exports.tintRed = shader(function() {/*

varying vec2 vTextureCoord;
uniform sampler2D uImage;

void main(void) {
    gl_FragColor = vec4(1.0, 0.0, 0.0, 0.5) + texture2D(uImage, vTextureCoord);
}

*/});

exports.solid = shader(function() {/*

uniform sampler2D uImage;

uniform vec3 uAmbientColor;
uniform vec3 uLightDirection;
uniform vec3 uLightPosition;
uniform vec3 uLightColor;
uniform bool uPerFragmentLighting;

varying vec2 vTextureCoord;

varying vec3 vLightWeighting;
varying vec4 vTransformedNormal;
varying vec4 vPosition;

void main(void) {
    vec4 color = texture2D(uImage, vTextureCoord) + vec4(0.5, 0.5, 0.8, 1);

    vec3 lightWeighting;
    if (uPerFragmentLighting) {
        vec3 lightDirection = normalize(uLightPosition - vPosition.xyz);
        float directionalLightWeighting =
            max(dot(normalize(vTransformedNormal.xyz), lightDirection), 0.0);
        lightWeighting = uAmbientColor + uLightColor * directionalLightWeighting;
    } else {
        lightWeighting = vLightWeighting;
    }

    gl_FragColor = vec4(color.rgb * lightWeighting, color.a);
}

*/});

exports.radialBlur = shader(function() {/*

varying vec2 vTextureCoord;
uniform sampler2D uImage;
uniform float sampleDist;
uniform float sampleStrength;

void main() {
    vec2 dir=vec2(0.5,0.5)-vTextureCoord;
    float dist=sqrt(dir.x*dir.x+dir.y*dir.y);
    dir=normalize(dir);
    float samples [10];
    samples [0]=-0.08;
    samples [1]=-0.05;
    samples [2]=-0.03;
    samples [3]=-0.02;
    samples [4]=-0.01;
    samples [5]=0.01;
    samples [6]=0.02;
    samples [7]=0.03;
    samples [8]=0.05;
    samples [9]=0.08;
    vec4 sample = texture2D(uImage, vTextureCoord);
    vec4 avg=sample;
    vec4 col;
    for (int i = 0; i < 10; i++) {
      vec2 tC=min(vec2(0.999,0.999),max(vec2(0.001,0.001),vTextureCoord+dir*samples[i]*sampleDist));

      col=texture2D(uImage,tC);
      float intensity=col.r+col.g+col.b;
      avg += col;  
    }
    avg /= 11.0;
    float t=dist*sampleStrength;
    t=clamp(t,0.0,1.0);
    gl_FragColor = mix(sample,avg,t);
}      

*/});

exports.gaussianBlur = shader(function() {/*

varying vec2 vTextureCoord;
uniform sampler2D uImage;

uniform float radius;
uniform vec2 blurMultiplyVec;

const float blurSize = 1.0/512.0;
const float pi = 3.14159265;

void main() {
    float sigma = radius / 6.0;
    
  // Incremental Gaussian Coefficent Calculation (See GPU Gems 3 pp. 877 - 889)
  vec3 incrementalGaussian;
  incrementalGaussian.x = 1.0 / (sqrt(2.0 * pi) * sigma);
  incrementalGaussian.y = exp(-0.5 / (sigma * sigma));
  incrementalGaussian.z = incrementalGaussian.y * incrementalGaussian.y;

  vec4 avgValue = vec4(0.0, 0.0, 0.0, 0.0);
  float coefficientSum = 0.0;

  // Take the central sample first...
  avgValue += texture2D(uImage, vTextureCoord.xy) * incrementalGaussian.x;
  coefficientSum += incrementalGaussian.x;
  incrementalGaussian.xy *= incrementalGaussian.yz;
  
  // Go through the remaining 8 vertical samples (4 on each side of the center)
  for (float i = 1.0; i <= 32.0; i++) { 
    if (i < radius) {
        avgValue += texture2D(uImage, vTextureCoord.xy - i * blurSize * 
                              blurMultiplyVec) * incrementalGaussian.x;         
        avgValue += texture2D(uImage, vTextureCoord.xy + i * blurSize * 
                              blurMultiplyVec) * incrementalGaussian.x;         
        coefficientSum += 2.0 * incrementalGaussian.x;
        incrementalGaussian.xy *= incrementalGaussian.yz;
    }
  }

  gl_FragColor = avgValue / coefficientSum;
}

*/});
